// myhello.cpp  ->  myhello.dll (msvc6 : dll example project, select c++:codegen:"multithreaded dll" and add luabin include&lib paths)
#include "SDL/SDL.h"
#include "SDL_audioin.h"
// win:vc6 : add SDL_audioin.lib SDL.lib   to linker (and paths accordingly)
#include "kiss_fft.h"
#include "kiss_fftr.h"
#include <limits.h>

#include <stdio.h>
extern "C" {
	#include "lua.h"
	#include "lauxlib.h"
}
#define PROJECT_TABLENAME "audioin"

#ifdef LUA_API
#undef LUA_API
#endif
#ifdef WIN32
#define LUA_API __declspec(dllexport)
#else
#define LUA_API
#endif

extern "C" {
	int LUA_API luaopen_audioin (lua_State *L);
}


#define EVENT_TICKS		10	//How many ticks it idles in the event loop

#define AUDIO_SAMPLERATE			44100
#define AUDIO_SAMPLES				(2*1024)
#define AUDIO_FORMAT				AUDIO_S16
//~ #define SAMPLES_PER_TICK			(AUDIO_SAMPLERATE/TICKS_PER_SECOND)
//~ #define FXP_TICKS_PER_SAMPLE		(TO_FXP(TICKS_PER_SECOND)/AUDIO_SAMPLERATE)
#define FFT_SAMPLE_EXPAND_FACTOR	1
#define FFT_NUM_SAMPLES				(AUDIO_SAMPLES)
#define FFT_SCALE					(1.0/8.0)

int __running=0;

int gSampleBuffer_NumSamples = 0;
int gSampleBuffer_ByteSize = 0;
Sint16* gSampleBuffer = 0;

// bigger recording buffer
SDL_mutex*	gLock_RecordBuffer = 0;
int			gRecordBuffer_NumSamples_Allocated = 0;
int			gRecordBuffer_NumSamples_Filled = 0;
Sint16*		gRecordBuffer = 0;
bool		gbRecordActive = false;

int MyMin (int a,int b) { return (a < b) ? a : b; }
int MyMax (int a,int b) { return (a > b) ? a : b; }

static int LuaIntDefault (lua_State *L,int i,int def);

// just make buffer big enough so nothing is lost if you cannot poll often enough. too large doesn't affect latency, can be popped before full
// recordbuffer is filled when one block of SAMPLE-BUFFER is full, this DOES affect latency
// RecordStart (num_samples=128*1024)  should be multiple of audio-in-buffer numsamples (given in init, mainly for life/spectrum)
static int L_RecordStart (lua_State *L) {
	if (gbRecordActive) { printf("audioin:RecordStart error : record already active\n"); exit(0); }
	SDL_mutexP(gLock_RecordBuffer);
	int num_samples = MyMax(1,LuaIntDefault(L,1,128*1024));
	gRecordBuffer_NumSamples_Filled = 0;
	gRecordBuffer_NumSamples_Allocated = num_samples;
	gRecordBuffer = (Sint16*)malloc(num_samples * sizeof(Sint16));
	gbRecordActive = true; 
	SDL_mutexV(gLock_RecordBuffer);
	return 0;
}

static int L_RecordStop (lua_State *L) { 
	if (!gbRecordActive) { printf("audioin:RecordStop error : record not active\n"); exit(0); }
	SDL_mutexP(gLock_RecordBuffer);
	gbRecordActive = false; 
	gRecordBuffer_NumSamples_Filled = 0;
	gRecordBuffer_NumSamples_Allocated = 0;
	free(gRecordBuffer); 
	gRecordBuffer = 0;
	SDL_mutexV(gLock_RecordBuffer);
	return 0;
}

// returns binary-string of record buffer contents, buffer is emptied
static int L_RecordPop (lua_State *L) { 
	SDL_mutexP(gLock_RecordBuffer);
	lua_pushlstring(L,(char*)gRecordBuffer,sizeof(Sint16)*gRecordBuffer_NumSamples_Filled);
	gRecordBuffer_NumSamples_Filled = 0;
	SDL_mutexV(gLock_RecordBuffer);
	return 0; 
}

void	InitSampleBuffer	(int num_samples) {
	gSampleBuffer_NumSamples = num_samples;
	gSampleBuffer_ByteSize = num_samples * sizeof(Sint16);
	gSampleBuffer = (Sint16*)malloc(gSampleBuffer_ByteSize);
}



void SoundInCallback(void *udata, Uint8 *stream, int len) {
	if (__running) {
		//~ Sint16 *sData=(Sint16 *)stream;
		//~ int num_samples = len/sizeof(Sint16);
		int copylen = MyMin(len,gSampleBuffer_ByteSize);
		int zerolen = gSampleBuffer_ByteSize - copylen;
		if (zerolen > 0) memset(((char*)gSampleBuffer)+copylen,0,zerolen);
		memcpy(gSampleBuffer,stream,copylen);
	}
	if (gbRecordActive) {
		SDL_mutexP(gLock_RecordBuffer);
		int num_samples = MyMin(gRecordBuffer_NumSamples_Allocated-gRecordBuffer_NumSamples_Filled,len/sizeof(Sint16));
		if (num_samples > 0) {
			memcpy(&gRecordBuffer[gRecordBuffer_NumSamples_Filled],stream,num_samples * sizeof(Sint16));
			gRecordBuffer_NumSamples_Filled += num_samples;
		}
		SDL_mutexV(gLock_RecordBuffer);
	}
}

class cFFTConfig { public:
	kiss_fftr_cfg		cfg;
	kiss_fft_cpx		*kout;
	kiss_fft_scalar		*tbuf;
	int nfft;
	long maxrange;
	long maxrange_half;
	float outvalscale;
	
	cFFTConfig () { _ResetZero(); }
	void	_ResetZero () { cfg	= 0; kout = 0; tbuf	= 0; nfft = 0; }
	void	Release () {
		// TODO : memleak : cfg ? no idea how to free
		kiss_fft_cleanup();
		KISS_FFT_FREE(tbuf);
		KISS_FFT_FREE(kout);
		_ResetZero();
	}
	void	Init (int nfft) { // nfft=BufferNumSamples
		this->nfft = nfft;
		cfg		= kiss_fftr_alloc(nfft , 0, NULL, NULL);
		tbuf	= (kiss_fft_scalar*)KISS_FFT_MALLOC(nfft * sizeof(kiss_fft_scalar));
		kout	= (kiss_fft_cpx*)KISS_FFT_MALLOC(nfft * sizeof(kiss_fft_cpx));
		#if FIXED_POINT==32
			maxrange = LONG_MAX;
		#else
			maxrange = SHRT_MAX;/* works fine for float too*/
		#endif
		maxrange_half = maxrange / 2;
		outvalscale = 1.0 / (float)maxrange;
	}
	
	static inline Sint32 Range(Sint32 max, Sint32 min, Sint16 val) {
		Sint32 tVal=((Sint32)val)+32768;
		tVal*=(max-min);
		tVal/=65536;
		return((Sint32)tVal);
	}
	
	void	Execute (Sint16* pSamples,int num_samples,int sample_stretch=1) { // sample_stretch seems to not play any role
		//~ assert(FFT_SAMPLE_EXPAND_FACTOR == 1); // interpolation ruins freq analysis !
		//~ if (FFT_SAMPLE_EXPAND_FACTOR != 1) { printf("FFT_SAMPLE_EXPAND_FACTOR must be 1!\n"); exit(0); return; }
		//~ for (int i=0;i<nfft;i++) tbuf[i] = (i<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[i]):0;
		//~ for (int i=0;i<nfft;i++) tbuf[i] = ((i/FFT_SAMPLE_EXPAND_FACTOR)<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[i/FFT_SAMPLE_EXPAND_FACTOR]):0;
		
		// TODO : interesting : psdpng.c
		// TODO : interesting : twotonetest.c

		if (sample_stretch == 1) {
			for (int i=0;i<nfft;i++) tbuf[i] = (i<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[i]):0;
		} else {
			// hardcopy "interpolation"
			//~ for (int i=0;i<nfft;i++) tbuf[i] = ((i/sample_stretch)<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[i/sample_stretch]):0;
			
			// linear interpolation
			for (int i=0;i<nfft;i++) {
				int n = i/sample_stretch;
				float a = ((n+0)<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[(n+0)]):0;
				float b = ((n+1)<num_samples)?Range(maxrange_half,-maxrange_half,pSamples[(n+1)]):0;
				float f = (float)(i%sample_stretch) / (float)sample_stretch;
				
				tbuf[i] = f*b+(1.0-f)*a;
			}
		}
		
		kiss_fftr(cfg, tbuf, kout); // FOURIER TRANSFORM HAPPENS HERE !!!! (output written to kout)
	}
	
	//~ inline float GetOut (const int i,const int a) { return (float)(a?kout[i].i:kout[i].r) / (float)maxrange; }
	inline float GetOut (const int i,const int a) { return GetOutVal(i); }
	inline float GetOutVal (const int i) { 
		float x = kout[i].r * outvalscale;
		float y = kout[i].i * outvalscale;
		return x*x+y*y;
	}
	
	//~ #define FFT_ASCI_WIDTH 80
	#define FFT_ASCI_WIDTH 200
	#define FFT_ASCI_HEIGHT 15
	void	AsciOut (int w,int h) { // print kout on console
		int a=0;
		//~ for (int a=0;a<2;++a) {
			int i;
			int startidx = 1;
			float mymin = GetOut(startidx,a);
			float mymax = GetOut(startidx,a);
			for (i=startidx;i<nfft;++i) {
				float f = GetOut(i,a);
				if (f < mymin) mymin = f;
				if (f > mymax) mymax = f;
			}
			float usedrange = mymax - mymin;
			if (usedrange <= 0) usedrange = 1.0;
			//~ float valstep = usedrange / h;
			float valstep = 0.4 / h;
			//~ float valadd = -mymin;
			float valadd = 0.01;
			
			for (int y=h-1;y>=0;--y) { // inverse
				float valcel = valstep*y;
				for (int x=0;x<w;++x) {
					float f = valadd + FFT_SCALE * ((x<nfft)?GetOut(x,a):0);
					printf((f >= valcel) ? "#" : ".");
				}
				printf("\n");
			}
			printf("bounds:%0.4f  %0.4f %s nfft=%d\n",mymin,mymax,a?"A=0":"A=1",(int)nfft);
		//~ }
	}
};

cFFTConfig gFFT;

void	LuaAudioIn_Quit	() {
	__running=0;
	SDL_CloseAudioIn();
	SDL_DestroyMutex(gLock_RecordBuffer);
	SDL_Quit();
	gFFT.Release();
}


void	LuaAudioIn_AsciSpectroLoop	() {
	if (__running) { printf("LuaAudioIn_AsciSpectroLoop:call init first\n"); return; }
	while (true) {
		gFFT.Execute(gSampleBuffer,gSampleBuffer_NumSamples,FFT_SAMPLE_EXPAND_FACTOR);
		gFFT.AsciOut(FFT_ASCI_WIDTH,FFT_ASCI_HEIGHT); printf("\n");
		SDL_Delay(EVENT_TICKS);
	}
}

// returns true on success
bool	LuaAudioIn_Init	(int num_samples_record=AUDIO_SAMPLES,int num_samples_fft=FFT_NUM_SAMPLES,int sample_rate=AUDIO_SAMPLERATE) {
	if (__running) { printf("LuaAudioIn_Init:cannot init twice\n"); exit(0); return false; }
	__running=1;
	SDL_Event event;
	SDL_AudioSpec expected,result;
	
	//~ printf("sdl-audio:init...\n");
	if (SDL_Init(SDL_INIT_AUDIO)<0) { printf("SDL_Init error\n"); exit(0); return false; }
	//~ printf("sdl-audio:ok\n");
	
	gLock_RecordBuffer = SDL_CreateMutex();
	
	if (SDL_InitAudioIn()<0) { printf("SDL_InitAudioIn error\n"); exit(0); return false; }
	//~ printf("sdl-audio-in:ok\n");
	
	expected.format=AUDIO_FORMAT;
	expected.freq=sample_rate;
	expected.callback=SoundInCallback;
	expected.samples=num_samples_record;
	expected.channels=1;
	expected.userdata=&result;
	InitSampleBuffer(expected.samples);

	if (SDL_OpenAudioIn(&expected,&result)<0) {
		fprintf(stderr,"Can't open audio input: %s\n",SDL_GetError());
		exit(0); 
		//~ SDL_Quit();
		return false;
	}

	//Start the audio
	SDL_PauseAudioIn(0);
	
	gFFT.Init(num_samples_fft);
	return true;
}

static int L_helloworld (lua_State *L) {
	printf("audioin:hello world!\n");
	return 0;
}

static int LuaIntDefault (lua_State *L,int i,int def) { return (lua_gettop(L) >= i && !lua_isnil(L,i)) ? luaL_checkint(L,i) : def; }
	
static int L_Sleep (lua_State *L) {
	SDL_Delay(LuaIntDefault(L,1,10));
	return 0;
}

// 1=pause 2=go
static int L_Pause (lua_State *L) {
	SDL_PauseAudioIn(LuaIntDefault(L,1,1));
	return 0;
}

static int L_AsciSpectroStep (lua_State *L) {
	gFFT.Execute(gSampleBuffer,gSampleBuffer_NumSamples,FFT_SAMPLE_EXPAND_FACTOR);
	gFFT.AsciOut(FFT_ASCI_WIDTH,FFT_ASCI_HEIGHT); printf("\n");
	return 0;
}

static int L_AsciSpectroLoop (lua_State *L) {
	// sdl tutorial : http://lazyfoo.net/SDL_tutorials/lesson01/linux/cli/index.php
	printf("audioin:AsciSpectroLoop\n");
	LuaAudioIn_AsciSpectroLoop();
	return 0;
}


// bSuccess Init(num_samples_record=2048,num_samples_fft=2048,sample_rate=44100)
static int L_Init (lua_State *L) {
	lua_pushboolean(L,LuaAudioIn_Init(LuaIntDefault(L,1,2048),LuaIntDefault(L,2,2048),LuaIntDefault(L,3,44100)));
	return 1;
}

// ExecuteFFT(expandfactor=1,numsamples=max) : returns table{fft results},firstval (=DC and Nyquist=nfft/2?)
static int L_ExecuteFFT (lua_State *L) {
	gFFT.Execute(gSampleBuffer,gSampleBuffer_NumSamples,LuaIntDefault(L,1,1));
	lua_newtable(L);
	int numsamples = LuaIntDefault(L,2,gFFT.nfft)+1;
	if (numsamples > gFFT.nfft)
		numsamples = gFFT.nfft;
	for (int i=1;i<numsamples;++i) {
		lua_pushnumber( L, gFFT.GetOutVal(i));
		lua_rawseti( L, -2, i );
	}
	lua_pushnumber( L, gFFT.GetOutVal(0));
	return 2;
}


static int L_Quit (lua_State *L) {
	LuaAudioIn_Quit();
	return 0;
}

int LUA_API luaopen_audioin (lua_State *L) {
	printf("luaopen_audioin\n");
	struct luaL_reg driver[] = {
		{"helloworld",		L_helloworld},		
		{"AsciSpectroLoop",	L_AsciSpectroLoop},		
		{"AsciSpectroStep",	L_AsciSpectroStep},		
		{"Pause",			L_Pause},		
		{"Sleep",			L_Sleep},		
		{"Init",			L_Init},		
		{"Quit",			L_Quit},		
		{"ExecuteFFT",		L_ExecuteFFT},	
		
		{"RecordStart",		L_RecordStart},		
		{"RecordStop",		L_RecordStop},		
		{"RecordPop",		L_RecordPop},		
		{NULL, NULL},
	};
	luaL_openlib (L, "audioin", driver, 0);
	return 1;
}

/* test.lua :
require("audioin")
audioin.helloworld() 
*/
